/// @file   AuthnOracle.h
/// @brief  Oracle Authn Implementation.
/// @author Alejandro Álvarez Ayllón <aalvarez@cern.ch>
#include <cstring>
#include <dmlite/cpp/utils/security.h>
#include "AuthnOracle.h"
#include "Queries.h"

using namespace dmlite;
using namespace oracle;


AuthnOracle::AuthnOracle(oracle::occi::ConnectionPool* pool,
                         oracle::occi::Connection* conn,
                         const std::string& mapfile,
                         bool hostDnIsRoot) throw(DmException):
  pool_(pool), conn_(conn), mapFile_(mapfile), hostDnIsRoot_(hostDnIsRoot)
{
  // Nothing
}
  


AuthnOracle::~AuthnOracle() throw(DmException)
{
  this->pool_->terminateConnection(this->conn_);
}



std::string AuthnOracle::getImplId(void) const throw()
{
  return std::string("AuthnOracle");
}



SecurityContext* AuthnOracle::createSecurityContext(const SecurityCredentials& cred) throw (DmException)
{
  UserInfo user;
  std::vector<GroupInfo> groups;

  this->getIdMap(cred.clientName, cred.fqans, &user, &groups);
  return new SecurityContext(cred, user, groups);
}



GroupInfo AuthnOracle::newGroup(const std::string& gname) throw (DmException)
{
  gid_t     gid;

  try {
    // Get the last gid, increment and update
    occi::Statement* gidStmt = getPreparedStatement(this->conn_, STMT_GET_UNIQ_GID_FOR_UPDATE);

    occi::ResultSet* rs = gidStmt->executeQuery();

    // Update the gid
    if (rs->next()) {
      gid = rs->getNumber(1);
      ++gid;

      occi::Statement* updateGidStmt = getPreparedStatement(this->conn_, STMT_UPDATE_UNIQ_GID);
      updateGidStmt->setNumber(1, gid);
      updateGidStmt->executeUpdate();

      this->conn_->terminateStatement(updateGidStmt);
    }
    // Couldn't get, so insert it instead
    else {
      gid = 1;
      occi::Statement* insertGidStmt = getPreparedStatement(this->conn_, STMT_INSERT_UNIQ_GID);
      insertGidStmt->setNumber(1, gid);
      insertGidStmt->executeUpdate();

      this->conn_->terminateStatement(insertGidStmt);
    }

    gidStmt->closeResultSet(rs);
    this->conn_->terminateStatement(gidStmt);

    // Insert the group
    occi::Statement* groupStmt = getPreparedStatement(this->conn_, STMT_INSERT_GROUP);

    groupStmt->setNumber(1, gid);
    groupStmt->setString(2, gname);
    groupStmt->setNumber(3, 0);

    groupStmt->executeUpdate();

    this->conn_->terminateStatement(groupStmt);

    // Commit
    this->conn_->commit();
  }
  catch (occi::SQLException& e) {
    this->conn_->rollback();
    throw DmException(DM_INTERNAL_ERROR, e.getMessage());
  }

  // Build and return the GroupInfo
  GroupInfo g;
  g.name      = gname;
  g["gid"]    = gid;
  g["banned"] = 0;

  return g;
}



GroupInfo AuthnOracle::getGroup(const std::string& groupName) throw (DmException)
{
  GroupInfo group;
  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_GET_GROUPINFO_BY_NAME);

  stmt->setString(1, groupName);
  occi::ResultSet* rs = stmt->executeQuery();

  if (!rs->next())
    throw DmException(DM_NO_SUCH_GROUP, "Group %s not found", groupName.c_str());

  group.name      = rs->getString(2);
  group["gid"]    = static_cast<gid_t>(rs->getNumber(1));
  group["banned"] = static_cast<int>(rs->getNumber(3));

  stmt->closeResultSet(rs);
  this->conn_->terminateStatement(stmt);

  return group;
}



GroupInfo AuthnOracle::getGroup(const std::string& key,
                                const boost::any& value) throw (DmException)
{
  if (key != "gid")
    throw DmException(DM_NOT_A_KEY,
                      "AuthnOracle does not support querying by %s",
                      key.c_str());
  
  gid_t gid = Extensible::anyToUnsigned(value);
  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_GET_GROUPINFO_BY_GID);
  
  stmt->setNumber(1, gid);
  occi::ResultSet* rs = stmt->executeQuery();
  
  if (!rs->next())
    throw DmException(DM_NO_SUCH_GROUP, "Group %u not found", gid);
  
  GroupInfo group;  
  group.name      = rs->getString(2);
  group["gid"]    = static_cast<gid_t>(rs->getNumber(1));
  group["banned"] = static_cast<int>(rs->getNumber(3));

  stmt->closeResultSet(rs);
  this->conn_->terminateStatement(stmt);

  return group;
}



void AuthnOracle::updateGroup(const GroupInfo& group) throw (DmException)
{
  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_UPDATE_GROUP);
  
  try {
    stmt->setInt(1, group.getLong("banned"));
    stmt->setString(2, group.name);
    stmt->executeUpdate();
    this->conn_->commit();
    this->conn_->terminateStatement(stmt);
  }
  catch (occi::SQLException& ex) {
    this->conn_->rollback();
    throw DmException(DM_INTERNAL_ERROR, ex.getMessage());
  }
}



void AuthnOracle::deleteGroup(const std::string& groupName) throw (DmException)
{
  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_DELETE_GROUP);
  
  try {
    stmt->setString(1, groupName);
    stmt->executeUpdate();
    this->conn_->commit();
    this->conn_->terminateStatement(stmt);
  }
  catch (occi::SQLException& ex) {
    this->conn_->rollback();
    throw DmException(DM_INTERNAL_ERROR, ex.getMessage());
  }
}



UserInfo AuthnOracle::newUser(const std::string& uname) throw (DmException)
{
  uid_t uid;
  
  try {
    // Get the last uid, increment and update
    occi::Statement* uidStmt = getPreparedStatement(this->conn_, STMT_GET_UNIQ_UID_FOR_UPDATE);

    occi::ResultSet* rs = uidStmt->executeQuery();

    // Update the uid
    if (rs->next()) {
      uid = rs->getNumber(1);
      ++uid;

      occi::Statement* updateUidStmt = getPreparedStatement(this->conn_, STMT_UPDATE_UNIQ_UID);
      updateUidStmt->setNumber(1, uid);
      updateUidStmt->executeUpdate();
      this->conn_->terminateStatement(updateUidStmt);
    }
    // Couldn't get, so insert it instead
    else {
      uid = 1;

      occi::Statement* insertUidStmt = getPreparedStatement(this->conn_, STMT_INSERT_UNIQ_UID);
      insertUidStmt->setNumber(1, uid);
      insertUidStmt->executeUpdate();
      this->conn_->terminateStatement(insertUidStmt);
    }

    uidStmt->closeResultSet(rs);
    this->conn_->terminateStatement(uidStmt);

    // Insert the user
    occi::Statement* userStmt = getPreparedStatement(this->conn_, STMT_INSERT_USER);

    userStmt->setNumber(1, uid);
    userStmt->setString(2, uname);
    userStmt->setString(3, "");
    userStmt->setNumber(4, 0);

    userStmt->executeUpdate();

    this->conn_->terminateStatement(userStmt);

    // Commit
    this->conn_->commit();
  }
  catch (occi::SQLException& e) {
    this->conn_->rollback();
    throw DmException(DM_INTERNAL_ERROR, e.getMessage());
  }

  // Build and return the UserInfo
  UserInfo u;
  u.name      = uname;
  u["uid"]    = uid;
  u["banned"] = 0;

  return u;
}



UserInfo AuthnOracle::getUser(const std::string& userName) throw (DmException)
{
  UserInfo  user;
  
  // If the username is the host DN, root!
  if (this->hostDnIsRoot_ && userName == dmlite::getHostDN()) {
    user.name      = userName;
    user["ca"]     = std::string();
    user["banned"] = 0;
    user["uid"]    = 0u;
  }
  else {
    occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_GET_USERINFO_BY_NAME);

    stmt->setString(1, userName);
    occi::ResultSet* rs = stmt->executeQuery();

    if (!rs->next())
      throw DmException(DM_NO_SUCH_USER, "User %s not found", userName.c_str());

    user["uid"]    = static_cast<uid_t>(rs->getNumber(1));
    user.name      = rs->getString(2);
    user["ca"]     = rs->getString(3);
    user["banned"] = static_cast<int>(rs->getNumber(4));

    stmt->closeResultSet(rs);
    this->conn_->terminateStatement(stmt);
  }

  return user;
}



UserInfo AuthnOracle::getUser(const std::string& key,
                              const boost::any& value) throw (DmException)
{
  if (key != "uid")
    throw DmException(DM_NOT_A_KEY,
                      "AuthnOracle does not support querying by %s",
                      key.c_str());
  
  uid_t uid = Extensible::anyToUnsigned(value);
  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_GET_USERINFO_BY_UID);

  stmt->setNumber(1, uid);
  occi::ResultSet* rs = stmt->executeQuery();
  
  if (!rs->next())
    throw DmException(DM_NO_SUCH_USER, "User %u not found", uid);
  
  UserInfo user;
  
  user["uid"]    = static_cast<uid_t>(rs->getNumber(1));
  user.name      = rs->getString(2);
  user["ca"]     = rs->getString(3);
  user["banned"] = static_cast<int>(rs->getNumber(4));
  
  return user;
}



void AuthnOracle::updateUser(const UserInfo& user) throw (DmException)
{
  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_UPDATE_USER);
  
  try {
    stmt->setInt(1, user.getLong("banned"));
    stmt->setString(2, user.name);
    stmt->executeUpdate();
    this->conn_->commit();
    this->conn_->terminateStatement(stmt);
  }
  catch (occi::SQLException& ex) {
    this->conn_->rollback();
    throw DmException(DM_INTERNAL_ERROR, ex.getMessage());
  }
}



void AuthnOracle::deleteUser(const std::string& userName) throw (DmException)
{
  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_DELETE_USER);
  
  try {
    stmt->setString(1, userName);
    stmt->executeUpdate();
    this->conn_->commit();
    this->conn_->terminateStatement(stmt);
  }
  catch (occi::SQLException& ex) {
    this->conn_->rollback();
    throw DmException(DM_INTERNAL_ERROR, ex.getMessage());
  }
}



std::vector<GroupInfo> AuthnOracle::getGroups(void) throw (DmException)
{
  std::vector<GroupInfo> groups;
  GroupInfo group;
  
  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_GET_ALL_GROUPS);
  occi::ResultSet* rs = stmt->executeQuery();
  
  while (rs->next()) {
    group["gid"]    = static_cast<gid_t>(rs->getNumber(1));
    group.name      = rs->getString(2);
    group["banned"] = static_cast<int>(rs->getNumber(3));
    
    groups.push_back(group);
  }
  
  stmt->closeResultSet(rs);
  this->conn_->terminateStatement(stmt);
  
  return groups;
}



std::vector<UserInfo> AuthnOracle::getUsers(void) throw (DmException)
{
  std::vector<UserInfo> users;
  UserInfo user;
  
  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_GET_ALL_USERS);
  occi::ResultSet* rs = stmt->executeQuery();
  
  while (rs->next()) {
    user["uid"]    = static_cast<uid_t>(rs->getNumber(1));
    user.name      = rs->getString(2);
    user["ca"]     = rs->getString(3);
    user["banned"] = static_cast<int>(rs->getNumber(4));
    
    users.push_back(user);
  }
  
  return users;
}



void AuthnOracle::getIdMap(const std::string& userName,
                                 const std::vector<std::string>& groupNames,
                                 UserInfo* user,
                                 std::vector<GroupInfo>* groups) throw (DmException)
{
  std::string vo;
  GroupInfo   group;

  // Clear
  groups->clear();

  // User mapping
  try {
    *user = this->getUser(userName);
  }
  catch (DmException& e) {
    if (e.code() == DM_NO_SUCH_USER)
      *user = this->newUser(userName);
    else
      throw;
  }

  // No VO information, so use the mapping file to get the group
  if (groupNames.empty()) {
    vo = voFromDn(this->mapFile_, userName);
    try {
      group = this->getGroup(vo);
    }
    catch (DmException& e) {
      if (e.code() == DM_NO_SUCH_GROUP)
        group = this->newGroup(vo);
      else
        throw;
    }
    groups->push_back(group);
  }
  else {
    // Get group info
    std::vector<std::string>::const_iterator i;
    for (i = groupNames.begin(); i != groupNames.end(); ++i) {
      vo = voFromRole(*i);
      try {
        group = this->getGroup(vo);
      }
      catch (DmException& e) {
        if (e.code() == DM_NO_SUCH_GROUP)
          group = this->newGroup(vo);
        else
          throw;
      }
      groups->push_back(group);
    }
  }
}
