/// @file   NsOracle.cpp
/// @brief  Oracle NS Implementation.
/// @author Alejandro Álvarez Ayllón <aalvarez@cern.ch>
#include "NsOracle.h"
#include <assert.h>
#include <cstring>
#include <cstdlib>
#include <dmlite/cpp/utils/urls.h>
#include "Queries.h"

using namespace dmlite;
using namespace oracle;


static void getMetadata(occi::ResultSet* rs, ExtendedStat* meta) throw(DmException)
{
  meta->stat.st_ino   = static_cast<long>(rs->getNumber(1));
  meta->parent        = static_cast<long>(rs->getNumber(2));

  meta->guid          = rs->getString(3);
  meta->name          = rs->getString(4);

  meta->stat.st_mode  = rs->getNumber( 5);
  meta->stat.st_nlink = rs->getNumber( 6);
  meta->stat.st_uid   = rs->getNumber( 7);
  meta->stat.st_gid   = rs->getNumber( 8);
  meta->stat.st_size  = static_cast<long>(rs->getNumber(9));
  meta->stat.st_atime = rs->getNumber(10);
  meta->stat.st_mtime = rs->getNumber(11);
  meta->stat.st_ctime = rs->getNumber(12);
  meta->status        = static_cast<ExtendedStat::FileStatus>(rs->getString(14)[0]);

  meta->csumtype      = rs->getString(15);
  meta->csumvalue     = rs->getString(16);
  
  meta->acl = Acl(rs->getString(17));
  
  (*meta)["type"] = rs->getNumber(13);
}



INodeOracle::INodeOracle(oracle::occi::ConnectionPool* pool,
                         oracle::occi::Connection* conn) throw (DmException):
  pool_(pool), conn_(conn), transactionLevel_(0)
{
  // Nothing
}



INodeOracle::~INodeOracle() throw (DmException)
{
  this->pool_->terminateConnection(this->conn_);
}



std::string INodeOracle::getImplId() const throw ()
{
  return std::string("INodeOracle");
}



void INodeOracle::setStackInstance(StackInstance* si) throw (DmException)
{
  // Nothing
}



void INodeOracle::setSecurityContext(const SecurityContext* ctx) throw (DmException)
{
  // Nothing
}



void INodeOracle::begin() throw (DmException)
{
  this->transactionLevel_++;
}



void INodeOracle::commit() throw (DmException)
{
  if (this->transactionLevel_ == 0)
    throw DmException(DM_INTERNAL_ERROR,
                      "INodeOracle::commit Inconsistent state (Maybe there is a\
 commit without a begin, or a badly handled error sequence.)");
  
  if (--this->transactionLevel_ == 0) {
    try {
      this->conn_->commit();
    }
    catch (occi::SQLException& ex) {
      throw DmException(DM_QUERY_FAILED, ex.getMessage());
    }
  }
}



void INodeOracle::rollback() throw (DmException)
{
  this->transactionLevel_ = 0;
  try {
    this->conn_->rollback();
  }
  catch (occi::SQLException& ex) {
    throw DmException(DM_QUERY_FAILED, ex.getMessage());
  }
}



ExtendedStat INodeOracle::create(const ExtendedStat& nf) throw (DmException)
{
  // Create the entry
  occi::Statement* fileStmt = getPreparedStatement(this->conn_, STMT_INSERT_FILE);
  
  this->begin();
  
  try {
    unsigned nlink = S_ISDIR(nf.stat.st_mode) ? 0 : 1;

    fileStmt->setNumber( 1, static_cast<long>(nf.parent));
    fileStmt->setString( 2, nf.name);
    fileStmt->setNumber( 3, nf.stat.st_mode);
    fileStmt->setNumber( 4, nlink);
    fileStmt->setNumber( 5, nf.stat.st_uid);
    fileStmt->setNumber( 6, nf.stat.st_gid);
    fileStmt->setNumber( 7, static_cast<long>(nf.stat.st_size));
    fileStmt->setNumber( 8, time(NULL));
    fileStmt->setNumber( 9, time(NULL));
    fileStmt->setNumber(10, time(NULL));
    fileStmt->setNumber(11, 0);
    fileStmt->setString(12, std::string(1, nf.status));
    fileStmt->setString(13, nf.csumtype);
    fileStmt->setString(14, nf.csumvalue);
    fileStmt->setString(15, nf.acl.serialize());

    fileStmt->executeUpdate();

    this->conn_->terminateStatement(fileStmt);

    // Increment the nlink
    this->updateNlink(nf.parent, +1);
  }
  catch (occi::SQLException& ex) {
    this->rollback();
    this->conn_->terminateStatement(fileStmt);
    throw DmException(DM_INTERNAL_ERROR, ex.getMessage());
  }
  
  // Commit and return back
  this->commit();
  return this->extendedStat(nf.parent, nf.name);
}



void INodeOracle::symlink(ino_t inode, const std::string& link) throw (DmException)
{
  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_INSERT_SYMLINK);

  this->begin();
  
  try {
    stmt->setNumber(1, static_cast<long>(inode));
    stmt->setString(2, link);

    stmt->executeUpdate();

    this->conn_->terminateStatement(stmt);
  }
  catch (occi::SQLException& ex) {
    this->rollback();
    this->conn_->terminateStatement(stmt);
    throw DmException(DM_INTERNAL_ERROR, ex.getMessage());
  }
  
  this->commit();
}



void INodeOracle::unlink(ino_t inode) throw (DmException)
{
  ExtendedStat file = this->extendedStat(inode);
  
  if (S_ISDIR(file.stat.st_mode) && file.stat.st_nlink > 0)
    throw DmException(DM_IS_DIRECTORY, "Inode %l is a directory and it is not empty", inode);

  
  this->begin();
  try {
    // Remove associated symlink
    occi::Statement* delSymlink = getPreparedStatement(this->conn_, STMT_DELETE_SYMLINK);
    delSymlink->setNumber(1, static_cast<long>(inode));
    delSymlink->executeUpdate();
    this->conn_->terminateStatement(delSymlink);

    // Remove associated comments
    occi::Statement* delComment = getPreparedStatement(this->conn_, STMT_DELETE_COMMENT);
    delComment->setNumber(1, static_cast<long>(inode));
    delComment->executeUpdate();
    this->conn_->terminateStatement(delComment);

    // Remove file itself
    occi::Statement* delFile = getPreparedStatement(this->conn_, STMT_DELETE_FILE);
    delFile->setNumber(1, static_cast<long>(inode));
    delFile->executeUpdate();
    this->conn_->terminateStatement(delFile);

    // And decrement nlink
    this->updateNlink(file.parent, -1);
  }
  catch (DmException& e) {
    this->rollback();
    throw;
  }
  catch (occi::SQLException& e) {
    this->rollback();
    throw DmException(DM_INTERNAL_ERROR, e.getMessage());
  }

  // Done!
  this->commit();
}


  
void INodeOracle::move(ino_t inode, ino_t dest) throw (DmException)
{
  occi::Statement* changeParentStmt = getPreparedStatement(this->conn_, STMT_CHANGE_PARENT);

  ExtendedStat file = this->extendedStat(inode);
  
  this->begin();
  try {
    changeParentStmt->setNumber(1, static_cast<long>(dest));
    changeParentStmt->setNumber(2, time(NULL));
    changeParentStmt->setNumber(3, static_cast<long>(inode));

    changeParentStmt->executeUpdate();

    this->conn_->terminateStatement(changeParentStmt);

    // Reduce nlinks from old
    this->updateNlink(file.parent, -1);

    // Increment from new
    this->updateNlink(dest, +1);
  }
  catch (occi::SQLException& e) {
    this->rollback();
    this->conn_->terminateStatement(changeParentStmt);
    throw DmException(DM_INTERNAL_ERROR, e.getMessage());
  }
  
  this->commit();
}



void INodeOracle::rename(ino_t inode, const std::string& name) throw (DmException)
{
  occi::Statement* changeNameStmt = getPreparedStatement(this->conn_, STMT_CHANGE_NAME);

  this->begin();
  try {
    changeNameStmt->setString(1, name);
    changeNameStmt->setNumber(2, time(NULL));
    changeNameStmt->setNumber(3, static_cast<long>(inode));

    changeNameStmt->executeUpdate();
    this->conn_->terminateStatement(changeNameStmt);
  }
  catch (occi::SQLException& e) {
    this->rollback();
    this->conn_->terminateStatement(changeNameStmt);
    throw DmException(DM_INTERNAL_ERROR, e.getMessage());
  }
  
  this->commit();
}



ExtendedStat INodeOracle::extendedStat(ino_t inode) throw (DmException)
{
  ExtendedStat     meta;
  occi::ResultSet* rs   = 0x00;
  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_GET_FILE_BY_ID);

  try {
    stmt->setNumber(1, static_cast<long>(inode));
    rs = stmt->executeQuery();
    if (!rs->next()) {
      stmt->closeResultSet(rs);
      this->conn_->terminateStatement(stmt);
      throw DmException(DM_NO_SUCH_FILE, "File %ld not found", inode);
    }
    getMetadata(rs, &meta);
  }
  catch (occi::SQLException& ex) {
    if (rs)
      stmt->closeResultSet(rs);
    this->conn_->terminateStatement(stmt);
    throw DmException(DM_QUERY_FAILED, ex.getMessage());
  }

  stmt->closeResultSet(rs);
  this->conn_->terminateStatement(stmt);
  return meta;
}



ExtendedStat INodeOracle::extendedStat(ino_t parent, const std::string& name) throw (DmException)
{
  ExtendedStat meta;
  occi::ResultSet* rs   = 0x00;
  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_GET_FILE_BY_NAME);

  try {
    stmt->setNumber(1, static_cast<long>(parent));
    stmt->setString(2, name);
    rs = stmt->executeQuery();
    if (!rs->next()) {
      stmt->closeResultSet(rs);
      this->conn_->terminateStatement(stmt);
      throw DmException(DM_NO_SUCH_FILE, name + " not found");
    }
    getMetadata(rs, &meta);
  }
  catch (occi::SQLException& ex) {
    if (rs)
      stmt->closeResultSet(rs);
    this->conn_->terminateStatement(stmt);
    throw DmException(DM_QUERY_FAILED, ex.getMessage());
  }

  stmt->closeResultSet(rs);
  this->conn_->terminateStatement(stmt);
  return meta;
}



ExtendedStat INodeOracle::extendedStat(const std::string& guid) throw (DmException)
{
  ExtendedStat     meta;
  occi::ResultSet* rs   = 0x00;
  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_GET_FILE_BY_GUID);

  try {
    stmt->setString(1, guid);
    rs = stmt->executeQuery();
    if (!rs->next()) {
      stmt->closeResultSet(rs);
      this->conn_->terminateStatement(stmt);
      throw DmException(DM_NO_SUCH_FILE, "File with guid " + guid + " not found");
    }
    getMetadata(rs, &meta);
  }
  catch (occi::SQLException& ex) {
    if (rs)
      stmt->closeResultSet(rs);
    this->conn_->terminateStatement(stmt);
    throw DmException(DM_QUERY_FAILED, ex.getMessage());
  }

  stmt->closeResultSet(rs);
  this->conn_->terminateStatement(stmt);
  return meta;
}



SymLink INodeOracle::readLink(ino_t inode) throw (DmException)
{
  SymLink link;
  occi::ResultSet* rs   = 0x00;
  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_GET_SYMLINK);

  try {
    stmt->setNumber(1, static_cast<long>(inode));
    rs = stmt->executeQuery();
    if (!rs->next()) {
      stmt->closeResultSet(rs);
      this->conn_->terminateStatement(stmt);
      throw DmException(DM_NO_SUCH_FILE, "Link %ld not found", inode);
    }
    link.inode = static_cast<long>(rs->getNumber(1));
    link.link  = rs->getString(2);
  }
  catch (occi::SQLException& ex) {
    if (rs)
      stmt->closeResultSet(rs);
    this->conn_->terminateStatement(stmt);
    throw DmException(DM_QUERY_FAILED, ex.getMessage());
  }

  stmt->closeResultSet(rs);
  this->conn_->terminateStatement(stmt);
  return link;
}



void INodeOracle::addReplica(const Replica& replica) throw (DmException)
{
  std::string  host;

  // Note: Rely on the Oracle constraint, and raise DM_EXISTS if violated
  // If server is empty, parse the surl
  if (replica.server.empty()) {
    Url u(replica.rfn);
    host = u.domain;
    if (host.empty())
      throw DmException(DM_INVALID_VALUE,
                       "Empty server specified, and SFN does not include it: %s",
                       replica.rfn.c_str());
  }
  else {
    host = replica.server;
  }

  // Add it
  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_ADD_REPLICA);
  
  this->begin();
  try {

    stmt->setNumber( 1, static_cast<long>(replica.fileid));
    stmt->setNumber( 2, time(NULL));
    stmt->setNumber( 3, time(NULL));
    stmt->setNumber( 4, time(NULL));
    stmt->setNumber( 5, time(NULL));

    stmt->setNull  ( 6, occi::OCCISTRING);
    stmt->setString( 7, std::string(1, replica.status));
    stmt->setString( 8, std::string(1, replica.type));
    stmt->setNull  ( 9, occi::OCCISTRING);
    stmt->setString(10, replica.getString("pool"));
    stmt->setString(11, host);
    stmt->setString(12, replica.getString("filesystem"));
    stmt->setString(13, replica.rfn);

    stmt->executeUpdate();

    this->conn_->terminateStatement(stmt);
  }
  catch (occi::SQLException& ex) {
    this->rollback();
    if (ex.getErrorCode() == 1)
      throw DmException(DM_EXISTS,
                        "Replica %s already registered", replica.rfn.c_str());
    else
      throw DmException(DM_QUERY_FAILED, ex.getMessage());
  }
  this->commit();
}



void INodeOracle::deleteReplica(const Replica& replica) throw (DmException)
{
  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_DELETE_REPLICA);
  
  this->begin();
  try {
    stmt->setNumber(1, static_cast<long>(replica.fileid));
    stmt->setString(2, replica.rfn);
  
    stmt->executeUpdate();
    this->conn_->terminateStatement(stmt);
  }
  catch (occi::SQLException& e) {
    this->rollback();
    this->conn_->terminateStatement(stmt);
    throw DmException(DM_INTERNAL_ERROR, e.getMessage());
  }
  this->commit();
}



std::vector<Replica> INodeOracle::getReplicas(ino_t inode) throw (DmException)
{
  Replica replica;

  // MySQL statement
  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_GET_FILE_REPLICAS);
  occi::ResultSet* rs;

  // Execute query
  try {
    stmt->setNumber(1, static_cast<long>(inode));
    rs = stmt->executeQuery();
  }
  catch (occi::SQLException& ex) {
    this->conn_->terminateStatement(stmt);
    throw DmException(DM_QUERY_FAILED, ex.getMessage());
  }

  // Bind result
  std::vector<Replica> replicas;

  while (rs->next()) {
    replica.clear();
    replica.replicaid  = static_cast<long>(rs->getNumber(1));
    replica.fileid     = static_cast<long>(rs->getNumber(2));
    replica.nbaccesses = static_cast<long>(rs->getNumber(3));
    replica.atime      = rs->getNumber(4);
    replica.ptime      = rs->getNumber(5);
    replica.ltime      = rs->getNumber(6);
    replica.status     = static_cast<Replica::ReplicaStatus>(rs->getString(7)[0]);
    replica.type       = static_cast<Replica::ReplicaType>(rs->getString(8)[0]);
    replica.server     = rs->getString(10);
    replica.rfn        = rs->getString(12);
    
    replica["pool"]       = rs->getString( 9);
    replica["filesystem"] = rs->getString(11);

    replicas.push_back(replica);
  }
  stmt->closeResultSet(rs);
  this->conn_->terminateStatement(stmt);

  return replicas;
}



Replica INodeOracle::getReplica(int64_t rid) throw (DmException)
{
  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_GET_REPLICA_BY_ID);
  stmt->setNumber(1, static_cast<long>(rid));

  occi::ResultSet* rs = stmt->executeQuery();

  if (!rs->next()) {
    this->conn_->terminateStatement(stmt);
    throw DmException(DM_NO_SUCH_REPLICA, "Replica %d not found", rid);
  }

  Replica r;

  r.replicaid  = static_cast<long>(rs->getNumber(1));
  r.fileid     = static_cast<long>(rs->getNumber(2));
  r.nbaccesses = static_cast<long>(rs->getNumber(3));
  r.atime      = rs->getNumber(4);
  r.ptime      = rs->getNumber(5);
  r.ltime      = rs->getNumber(6);
  r.status     = static_cast<Replica::ReplicaStatus>(rs->getString(7)[0]);
  r.type       = static_cast<Replica::ReplicaType>(rs->getString(8)[0]);
  r.server     = rs->getString(10);
  r.rfn        = rs->getString(12);
  
  r["pool"]       = rs->getString( 9);
  r["filesystem"] = rs->getString(11);

  stmt->closeResultSet(rs);
  this->conn_->terminateStatement(stmt);
  return r; 
}



Replica INodeOracle::getReplica(const std::string& sfn) throw (DmException)
{
  occi::Statement* stmt = getPreparedStatement(this->conn_,
                                               STMT_GET_REPLICA_BY_URL);
  stmt->setString(1, sfn);

  occi::ResultSet* rs = stmt->executeQuery();

  if (!rs->next()) {
    this->conn_->terminateStatement(stmt);
    throw DmException(DM_NO_SUCH_REPLICA, "Replica %s not found", sfn.c_str());
  }

  Replica r;

  r.replicaid  = static_cast<long>(rs->getNumber(1));
  r.fileid     = static_cast<long>(rs->getNumber(2));
  r.nbaccesses = static_cast<long>(rs->getNumber(3));
  r.atime      = rs->getNumber(4);
  r.ptime      = rs->getNumber(5);
  r.ltime      = rs->getNumber(6);
  r.status     = static_cast<Replica::ReplicaStatus>(rs->getString(7)[0]);
  r.type       = static_cast<Replica::ReplicaType>(rs->getString(8)[0]);
  r.server     = rs->getString(10);
  r.rfn        = rs->getString(12);
  
  r["pool"]       = rs->getString( 9);
  r["filesystem"] = rs->getString(11);  

  stmt->closeResultSet(rs);
  this->conn_->terminateStatement(stmt);
  return r;
}



void INodeOracle::updateReplica(const Replica& replica) throw (DmException)
{
  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_UPDATE_REPLICA);
  stmt->setNumber(1, replica.atime);
  stmt->setNumber(2, replica.ltime);
  stmt->setNumber(3, static_cast<long>(replica.nbaccesses));
  stmt->setString(4, std::string(1, replica.status));
  stmt->setString(5, std::string(1, replica.type));
  stmt->setString(6, replica.rfn);

  this->begin();
  try {
    unsigned affected =  stmt->executeUpdate();
    this->conn_->terminateStatement(stmt);
    if (affected == 0) {
      this->rollback();
      throw DmException(DM_NO_SUCH_REPLICA,
                        "Replica '%s' not found", replica.rfn.c_str());
    }
  }
  catch (occi::SQLException& e) {
    this->rollback();
    this->conn_->terminateStatement(stmt);
    throw DmException(DM_INTERNAL_ERROR, e.getMessage());
  }
  this->commit();
}



void INodeOracle::utime(ino_t inode, const struct utimbuf* buf) throw (DmException)
{
  // If NULL, current time
  struct utimbuf internal;
  if (buf == 0x00) {
    buf = &internal;
    internal.actime  = time(NULL);
    internal.modtime = time(NULL);
  }

  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_UTIME);
  this->begin();
  try {
    stmt->setNumber(1, buf->actime);
    stmt->setNumber(2, buf->modtime);
    stmt->setNumber(3, time(NULL));
    stmt->setNumber(4, static_cast<long>(inode));

    stmt->executeUpdate();
    this->conn_->terminateStatement(stmt);
  }
  catch (occi::SQLException& e) {
    this->rollback();
    this->conn_->terminateStatement(stmt);
    throw DmException(DM_INTERNAL_ERROR, e.getMessage());
  }
  
  this->commit();
}



void INodeOracle::setMode(ino_t inode, uid_t uid, gid_t gid,
                          mode_t mode, const Acl& acl) throw (DmException)
{
  ExtendedStat xs = this->extendedStat(inode);
  
  // Make sure values are those we want
  const Acl* finalAcl;
  if (uid == -1) uid = xs.stat.st_uid;
  if (gid == -1) gid = xs.stat.st_gid;
  if (acl.empty())
    finalAcl = &xs.acl;
  else
    finalAcl = &acl;
  
  mode = (xs.stat.st_mode & S_IFMT) | (mode & ~S_IFMT);
  
  // Perform
  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_UPDATE_PERMS);
  
  this->begin();
  try {
    stmt->setNumber(1, uid);
    stmt->setNumber(2, gid);
    stmt->setNumber(3, mode);
    stmt->setString(4, finalAcl->serialize());
    stmt->setNumber(5, time(NULL));
    stmt->setNumber(6, static_cast<long>(inode));
    stmt->executeUpdate();
    this->conn_->terminateStatement(stmt);
  }
  catch (occi::SQLException& e) {
    this->rollback();
    this->conn_->terminateStatement(stmt);
    throw DmException(DM_INTERNAL_ERROR, e.getMessage());
  }
  
  this->commit();
}



void INodeOracle::setSize(ino_t inode, size_t size) throw (DmException)
{
  throw DmException(DM_NOT_IMPLEMENTED, "changeSize not implemented in Oracle");
}



void INodeOracle::setChecksum(ino_t inode, const std::string& csumtype,
                                 const std::string& csumvalue) throw (DmException)
{
  throw DmException(DM_NOT_IMPLEMENTED, "changeChecksum not implemented in Oracle");
}



std::string INodeOracle::getComment(ino_t inode) throw (DmException)
{
  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_GET_COMMENT);

  stmt->setNumber(1, static_cast<long>(inode));
  occi::ResultSet* rs = stmt->executeQuery();

  if (!rs->next())
    throw DmException(DM_NO_COMMENT, "There is no comment for " + inode);

  std::string comment = rs->getString(1);
  
  stmt->closeResultSet(rs);
  this->conn_->terminateStatement(stmt);

  // Done here!
  return comment;
}



void INodeOracle::setComment(ino_t inode, const std::string& comment) throw (DmException)
{
  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_SET_COMMENT);  
  
  this->begin();

  try {
    stmt->setString(1, comment);
    stmt->setNumber(2, static_cast<long>(inode));

    if (stmt->executeUpdate() == 0) {
      // No update! Try inserting
      occi::Statement* stmti = getPreparedStatement(this->conn_, STMT_INSERT_COMMENT);

      stmti->setNumber(1, static_cast<long>(inode));
      stmti->setString(2, comment);

      stmti->executeUpdate();
    }

    this->conn_->terminateStatement(stmt);
  }
  catch (occi::SQLException& e) {
    this->rollback();
    this->conn_->terminateStatement(stmt);
    throw DmException(DM_INTERNAL_ERROR, e.getMessage());
  }
  
  this->commit();
}



void INodeOracle::deleteComment(ino_t inode) throw (DmException)
{
  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_DELETE_COMMENT);
  
  this->begin();

  try {
    stmt->setNumber(1, static_cast<long>(inode));

    stmt->executeUpdate();

    this->conn_->terminateStatement(stmt);
  }
  catch (occi::SQLException& e) {
    this->rollback();
    this->conn_->terminateStatement(stmt);
    throw DmException(DM_INTERNAL_ERROR, e.getMessage());
  }
  
  this->commit();
}



void INodeOracle::setGuid(ino_t inode, const std::string& guid) throw (DmException)
{
  occi::Statement* stmt = getPreparedStatement(this->conn_, STMT_SET_GUID);
  
  this->begin();

  try {
    stmt->setString(1, guid);
    stmt->setNumber(2, static_cast<long>(inode));

    stmt->executeUpdate();

    this->conn_->terminateStatement(stmt);
  }
  catch (occi::SQLException& e) {
    this->rollback();
    this->conn_->terminateStatement(stmt);
    throw DmException(DM_INTERNAL_ERROR, e.getMessage());
  }
  
  this->commit();
}



void INodeOracle::updateExtendedAttributes(ino_t inode,
                                           const Extensible& attr) throw (DmException)
{
  throw DmException(DM_NOT_IMPLEMENTED,
                    "INodeOracle does not implement extended attributes");
}



IDirectory* INodeOracle::openDir(ino_t inode) throw (DmException)
{
  NsOracleDir  *dir;

  dir = new NsOracleDir();
  dir->dirId = inode;
  dir->stmt = getPreparedStatement(this->conn_, STMT_GET_LIST_FILES);

  try {
    dir->stmt->setNumber(1, static_cast<long>(dir->dirId));
    dir->rs = dir->stmt->executeQuery();
    return dir;
  }
  catch (occi::SQLException& e) {
    delete dir;
    throw DmException(DM_INTERNAL_ERROR, e.getMessage());
  }
}



void INodeOracle::closeDir(IDirectory* dir) throw (DmException)
{
  NsOracleDir *dirp;

  if (dir == 0x00)
    throw DmException(DM_NULL_POINTER, std::string("Tried to close a null dir"));

  dirp = dynamic_cast<NsOracleDir*>(dir);

  dirp->stmt->closeResultSet(dirp->rs);
  this->conn_->terminateStatement(dirp->stmt);
  delete dirp;
}



ExtendedStat* INodeOracle::readDirx(IDirectory* dir) throw (DmException)
{
  NsOracleDir *dirp;

  if (dir == 0x00)
    throw DmException(DM_NULL_POINTER, "Tried to read a null dir");

  dirp = dynamic_cast<NsOracleDir*>(dir);

  if (dirp->rs->next()) {
    getMetadata(dirp->rs, &dirp->current);
    memset(&dirp->ds, 0x00, sizeof(struct dirent));
    dirp->ds.d_ino  = dirp->current.stat.st_ino;
    strncpy(dirp->ds.d_name,
            dirp->current.name.c_str(),
            sizeof(dirp->ds.d_name));


    return &dirp->current;
  }
  else {
    return 0x00;
  }
}



struct dirent* INodeOracle::readDir(IDirectory* dir) throw (DmException)
{
  if (this->readDirx(dir) == 0)
    return 0x00;
  else
    return &(((NsOracleDir*)dir)->ds);
}



void INodeOracle::updateNlink(ino_t inode, int diff) throw (DmException)
{
  // inode 0 is reserved for / parent
  if (inode == 0)
    return;
  
  occi::Statement* nlinkStmt = getPreparedStatement(this->conn_, STMT_SELECT_NLINK_FOR_UPDATE);
  nlinkStmt->setNumber(1, static_cast<long>(inode));
  occi::ResultSet* nlinkRs = nlinkStmt->executeQuery();
  nlinkRs->next();
  long nlink = nlinkRs->getNumber(1);

  nlinkStmt->closeResultSet(nlinkRs);
  this->conn_->terminateStatement(nlinkStmt);

  occi::Statement* nlinkUpdateStmt = getPreparedStatement(this->conn_, STMT_UPDATE_NLINK);

  nlink += diff;
  nlinkUpdateStmt->setNumber(1, nlink);
  nlinkUpdateStmt->setNumber(2, time(NULL));
  nlinkUpdateStmt->setNumber(3, time(NULL));
  nlinkUpdateStmt->setNumber(4, static_cast<long>(inode));

  nlinkUpdateStmt->executeUpdate();
  this->conn_->terminateStatement(nlinkUpdateStmt);
}
